@@ -16,10 +16,11 @@ use cargo::util::command_prelude::{ArgMatches, ArgMatchesExt, CompileMode, Profi
 use cargo::{CliError, CliResult, Config};
 
 use anyhow::Error;
+use cargo_util::paths::copy;
 use semver::Version;
 
 use crate::build_targets::BuildTargets;
-use crate::install::{InstallPaths, LibType, UnixLibNames};
+use crate::install::InstallPaths;
 use crate::pkg_config_gen::PkgConfig;
 use crate::target;
 
@@ -60,23 +61,20 @@ fn build_include_file(
     Ok(())
 }
 
-/// Copy the pre-built C header from the asset directory
+/// Copy the pre-built C header from the asset directory to the root_dir
 fn copy_prebuilt_include_file(
     ws: &Workspace,
-    name: &str,
+    build_targets: &BuildTargets,
     root_output: &Path,
-    root_path: &Path,
 ) -> anyhow::Result<()> {
     ws.config()
         .shell()
-        .status("Building", "pre-built header file")?;
-    let mut header_name = PathBuf::from(name);
-    header_name.set_extension("h");
-
-    let source_path = root_path.join("assets").join(&header_name);
-    let target_path = root_output.join(header_name);
-
-    std::fs::copy(source_path, target_path)?;
+        .status("Copying", "pre-built header files")?;
+    for (from, to) in build_targets.extra.include.iter() {
+        let to = root_output.join(to);
+        std::fs::create_dir_all(to.parent().unwrap())?;
+        copy(from, to)?;
+    }
 
     Ok(())
 }
@@ -95,15 +93,15 @@ fn build_pc_file(name: &str, root_output: &Path, pc: &PkgConfig) -> anyhow::Resu
 
 fn build_pc_files(
     ws: &Workspace,
-    name: &str,
+    filename: &str,
     root_output: &Path,
     pc: &PkgConfig,
 ) -> anyhow::Result<()> {
     ws.config().shell().status("Building", "pkg-config files")?;
-    build_pc_file(name, root_output, pc)?;
+    build_pc_file(filename, root_output, pc)?;
     let pc_uninstalled = pc.uninstalled(&root_output);
     build_pc_file(
-        &format!("{}-uninstalled", name),
+        &format!("{}-uninstalled", filename),
         root_output,
         &pc_uninstalled,
     )
@@ -179,7 +177,7 @@ fn build_def_file(
 
         dumpbin
             .arg("/EXPORTS")
-            .arg(targetdir.join(format!("{}.dll", name)));
+            .arg(targetdir.join(format!("{}.dll", name.replace("-", "_"))));
         dumpbin.arg(format!("/OUT:{}", txt_path.to_str().unwrap()));
 
         let out = dumpbin.output()?;
@@ -235,34 +233,65 @@ fn build_implib_file(
     let os = &target.os;
     let env = &target.env;
 
-    if os == "windows" && env == "gnu" {
-        ws.config()
-            .shell()
-            .status("Building", "implib using dlltool")?;
-
+    if os == "windows" {
         let arch = &target.arch;
+        if env == "gnu" {
+            ws.config()
+                .shell()
+                .status("Building", "implib using dlltool")?;
 
-        let binutils_arch = match arch.as_str() {
-            "x86_64" => "i386:x86-64",
-            "x86" => "i386",
-            _ => unimplemented!("Windows support for {} is not implemented yet.", arch),
-        };
+            let binutils_arch = match arch.as_str() {
+                "x86_64" => "i386:x86-64",
+                "x86" => "i386",
+                _ => unimplemented!("Windows support for {} is not implemented yet.", arch),
+            };
 
-        let mut dlltool_command = std::process::Command::new(dlltool.to_str().unwrap_or("dlltool"));
-        dlltool_command.arg("-m").arg(binutils_arch);
-        dlltool_command.arg("-D").arg(format!("{}.dll", name));
-        dlltool_command
-            .arg("-l")
-            .arg(targetdir.join(format!("{}.dll.a", name)));
-        dlltool_command
-            .arg("-d")
-            .arg(targetdir.join(format!("{}.def", name)));
+            let mut dlltool_command =
+                std::process::Command::new(dlltool.to_str().unwrap_or("dlltool"));
+            dlltool_command.arg("-m").arg(binutils_arch);
+            dlltool_command.arg("-D").arg(format!("{}.dll", name));
+            dlltool_command
+                .arg("-l")
+                .arg(targetdir.join(format!("{}.dll.a", name)));
+            dlltool_command
+                .arg("-d")
+                .arg(targetdir.join(format!("{}.def", name)));
 
-        let out = dlltool_command.output()?;
-        if out.status.success() {
-            Ok(())
+            let out = dlltool_command.output()?;
+            if out.status.success() {
+                Ok(())
+            } else {
+                Err(anyhow::anyhow!("Command failed {:?}", dlltool_command))
+            }
         } else {
-            Err(anyhow::anyhow!("Command failed {:?}", dlltool_command))
+            ws.config().shell().status("Building", "implib using lib")?;
+            let target_str = format!("{}-pc-windows-msvc", &target.arch);
+            let mut lib = match cc::windows_registry::find(&target_str, "lib.exe") {
+                Some(command) => command,
+                None => std::process::Command::new("lib"),
+            };
+            let lib_arch = match arch.as_str() {
+                "x86_64" => "X64",
+                "x86" => "IX86",
+                _ => unimplemented!("Windows support for {} is not implemented yet.", arch),
+            };
+            lib.arg(format!(
+                "/DEF:{}",
+                targetdir.join(format!("{}.def", name)).display()
+            ));
+            lib.arg(format!("/MACHINE:{}", lib_arch));
+            lib.arg(format!("/NAME:{}.dll", name));
+            lib.arg(format!(
+                "/OUT:{}",
+                targetdir.join(format!("{}.dll.lib", name)).display()
+            ));
+
+            let out = lib.output()?;
+            if out.status.success() {
+                Ok(())
+            } else {
+                Err(anyhow::anyhow!("Command failed {:?}", lib))
+            }
         }
     } else {
         Ok(())
@@ -272,7 +301,7 @@ fn build_implib_file(
 struct FingerPrint<'a> {
     name: &'a str,
     root_output: &'a Path,
-    build_targets: &'a BuildTargets,
+    build_targets: BuildTargets,
     install_paths: &'a InstallPaths,
     static_libs: &'a str,
 }
@@ -287,7 +316,7 @@ impl<'a> FingerPrint<'a> {
     fn new(
         name: &'a str,
         root_output: &'a Path,
-        build_targets: &'a BuildTargets,
+        build_targets: BuildTargets,
         install_paths: &'a InstallPaths,
     ) -> Self {
         Self {
@@ -373,21 +402,24 @@ pub struct CApiConfig {
     pub header: HeaderCApiConfig,
     pub pkg_config: PkgConfigCApiConfig,
     pub library: LibraryCApiConfig,
+    pub install: InstallCApiConfig,
 }
 
 pub struct HeaderCApiConfig {
     pub name: String,
-    pub subdirectory: bool,
+    pub subdirectory: String,
     pub generation: bool,
     pub enabled: bool,
 }
 
 pub struct PkgConfigCApiConfig {
     pub name: String,
+    pub filename: String,
     pub description: String,
     pub version: String,
     pub requires: Option<String>,
     pub requires_private: Option<String>,
+    pub strip_include_path_components: usize,
 }
 
 pub struct LibraryCApiConfig {
@@ -398,6 +430,77 @@ pub struct LibraryCApiConfig {
     pub rustflags: Vec<String>,
 }
 
+pub struct InstallCApiConfig {
+    pub include: Vec<InstallTarget>,
+}
+
+pub enum InstallTarget {
+    Asset(InstallTargetPaths),
+    Generated(InstallTargetPaths),
+}
+
+#[derive(Clone)]
+pub struct InstallTargetPaths {
+    /// pattern to feed to glob::glob()
+    ///
+    /// if the InstallTarget is Asset its root is the the root_path
+    /// if the InstallTarget is Generated its root is the root_output
+    pub from: String,
+    /// The path to be joined to the canonical directory to install the files discovered by the
+    /// glob, e.g. `{includedir}/{to}` for includes.
+    pub to: String,
+}
+
+impl InstallTargetPaths {
+    pub fn from_value(value: &toml::value::Value, default_to: &str) -> anyhow::Result<Self> {
+        let from = value
+            .get("from")
+            .and_then(|v| v.as_str())
+            .ok_or_else(|| anyhow::anyhow!("a from field is required"))?;
+        let to = value
+            .get("to")
+            .and_then(|v| v.as_str())
+            .unwrap_or(default_to);
+
+        Ok(InstallTargetPaths {
+            from: from.to_string(),
+            to: to.to_string(),
+        })
+    }
+
+    pub fn install_paths(
+        &self,
+        root: &Path,
+    ) -> anyhow::Result<impl Iterator<Item = (PathBuf, PathBuf)>> {
+        let pattern = root.join(&self.from);
+        let base_pattern = if self.from.contains("/**") {
+            pattern
+                .iter()
+                .take_while(|&c| c != std::ffi::OsStr::new("**"))
+                .collect()
+        } else {
+            pattern.parent().unwrap().to_path_buf()
+        };
+        let pattern = pattern.to_str().unwrap();
+        let to = PathBuf::from(&self.to);
+        let g = glob::glob(pattern)?.filter_map(move |p| {
+            if let Ok(p) = p {
+                if p.is_file() {
+                    let from = p;
+                    let to = to.join(from.strip_prefix(&base_pattern).unwrap());
+                    Some((from, to))
+                } else {
+                    None
+                }
+            } else {
+                None
+            }
+        });
+
+        Ok(g)
+    }
+}
+
 fn load_manifest_capi_config(
     name: &str,
     root_path: &Path,
@@ -432,6 +535,22 @@ fn load_manifest_capi_config(
 
     let header = capi.and_then(|v| v.get("header"));
 
+    let subdirectory = header
+        .as_ref()
+        .and_then(|h| h.get("subdirectory"))
+        .map(|v| {
+            if let Ok(b) = v.clone().try_into::<bool>() {
+                Ok(if b {
+                    String::from(name)
+                } else {
+                    String::from("")
+                })
+            } else {
+                v.clone().try_into::<String>()
+            }
+        })
+        .unwrap_or_else(|| Ok(String::from(name)))?;
+
     let header = if let Some(ref capi) = capi {
         HeaderCApiConfig {
             name: header
@@ -440,11 +559,7 @@ fn load_manifest_capi_config(
                 .or_else(|| capi.get("header_name"))
                 .map(|v| v.clone().try_into())
                 .unwrap_or_else(|| Ok(String::from(name)))?,
-            subdirectory: header
-                .as_ref()
-                .and_then(|h| h.get("subdirectory"))
-                .map(|v| v.clone().try_into())
-                .unwrap_or(Ok(true))?,
+            subdirectory,
             generation: header
                 .as_ref()
                 .and_then(|h| h.get("generation"))
@@ -459,7 +574,7 @@ fn load_manifest_capi_config(
     } else {
         HeaderCApiConfig {
             name: String::from(name),
-            subdirectory: true,
+            subdirectory: String::from(name),
             generation: true,
             enabled: true,
         }
@@ -468,6 +583,7 @@ fn load_manifest_capi_config(
     let pc = capi.and_then(|v| v.get("pkg_config"));
     let pkg = ws.current().unwrap();
     let mut pc_name = String::from(name);
+    let mut pc_filename = String::from(name);
     let mut description = String::from(
         pkg.manifest()
             .metadata()
@@ -478,11 +594,15 @@ fn load_manifest_capi_config(
     let mut version = pkg.version().to_string();
     let mut requires = None;
     let mut requires_private = None;
+    let mut strip_include_path_components = 0;
 
     if let Some(ref pc) = pc {
         if let Some(override_name) = pc.get("name").and_then(|v| v.as_str()) {
             pc_name = String::from(override_name);
         }
+        if let Some(override_filename) = pc.get("filename").and_then(|v| v.as_str()) {
+            pc_filename = String::from(override_filename);
+        }
         if let Some(override_description) = pc.get("description").and_then(|v| v.as_str()) {
             description = String::from(override_description);
         }
@@ -495,14 +615,20 @@ fn load_manifest_capi_config(
         if let Some(req) = pc.get("requires_private").and_then(|v| v.as_str()) {
             requires_private = Some(String::from(req));
         }
+        strip_include_path_components = pc
+            .get("strip_include_path_components")
+            .map(|v| v.clone().try_into())
+            .unwrap_or_else(|| Ok(0))?
     }
 
     let pkg_config = PkgConfigCApiConfig {
         name: pc_name,
+        filename: pc_filename,
         description,
         version,
         requires,
         requires_private,
+        strip_include_path_components,
     };
 
     let library = capi.and_then(|v| v.get("library"));
@@ -544,10 +670,55 @@ fn load_manifest_capi_config(
         rustflags,
     };
 
+    let default_assets_include = InstallTargetPaths {
+        from: "assets/capi/include/**/*".to_string(),
+        to: header.subdirectory.clone(),
+    };
+
+    let default_legacy_asset_include = InstallTargetPaths {
+        from: format!("assets/{}.h", header.name),
+        to: header.subdirectory.clone(),
+    };
+
+    let default_generated_include = InstallTargetPaths {
+        from: "capi/include/**/*".to_string(),
+        to: header.subdirectory.clone(),
+    };
+
+    let mut include_targets = vec![
+        InstallTarget::Asset(default_assets_include),
+        InstallTarget::Asset(default_legacy_asset_include),
+        InstallTarget::Generated(default_generated_include),
+    ];
+
+    let install = capi.and_then(|v| v.get("install"));
+    if let Some(ref install) = install {
+        if let Some(includes) = install.get("include") {
+            if let Some(assets) = includes.get("asset").and_then(|v| v.as_array()) {
+                for asset in assets {
+                    let target_paths = InstallTargetPaths::from_value(asset, &header.subdirectory)?;
+                    include_targets.push(InstallTarget::Asset(target_paths));
+                }
+            }
+
+            if let Some(generated) = includes.get("generated").and_then(|v| v.as_array()) {
+                for gen in generated {
+                    let target_paths = InstallTargetPaths::from_value(gen, &header.subdirectory)?;
+                    include_targets.push(InstallTarget::Generated(target_paths));
+                }
+            }
+        }
+    }
+
+    let install = InstallCApiConfig {
+        include: include_targets,
+    };
+
     Ok(CApiConfig {
         header,
         pkg_config,
         library,
+        install,
     })
 }
 
@@ -642,7 +813,7 @@ impl Executor for Exec {
     }
 }
 
-use cargo::core::compiler::{unit_graph, Compilation, UnitInterner};
+use cargo::core::compiler::{unit_graph, UnitInterner};
 use cargo::ops::create_bcx;
 use cargo::util::profile;
 
@@ -668,7 +839,7 @@ fn compile_with_exec<'a>(
     exec: &Arc<dyn Executor>,
     leaf_args: &[String],
     global_args: &[String],
-) -> CargoResult<Compilation<'a>> {
+) -> CargoResult<Option<PathBuf>> {
     ws.emit_warnings()?;
     let interner = UnitInterner::new();
     let mut bcx = create_bcx(ws, options, &interner)?;
@@ -684,11 +855,32 @@ fn compile_with_exec<'a>(
 
     if options.build_config.unit_graph {
         unit_graph::emit_serialized_unit_graph(&bcx.roots, &bcx.unit_graph, ws.config())?;
-        return Compilation::new(&bcx);
+        return Ok(None);
     }
     let _p = profile::start("compiling");
     let cx = cargo::core::compiler::Context::new(&bcx)?;
-    cx.compile(exec)
+
+    let r = cx.compile(exec)?;
+
+    // NOTE: right now we have only a single cdylib since we have a single package to build.
+    let out_dir = r
+        .cdylibs
+        .iter()
+        .filter_map(|l| {
+            l.script_meta.map(|m| {
+                r.extra_env.get(&m).unwrap().iter().filter_map(|e| {
+                    if e.0 == "OUT_DIR" {
+                        Some(PathBuf::from(&e.1))
+                    } else {
+                        None
+                    }
+                })
+            })
+        })
+        .flatten()
+        .next();
+
+    Ok(out_dir)
 }
 
 pub fn cbuild(
@@ -772,10 +964,11 @@ pub fn cbuild(
         leaf_args.push("target-feature=+crt-static".into());
     }
 
-    let build_targets =
-        BuildTargets::new(&name, &rustc_target, &root_output, &libkinds, &capi_config);
+    let mut build_targets =
+        BuildTargets::new(&name, &rustc_target, &root_output, &libkinds, &capi_config)?;
 
-    let mut finger_print = FingerPrint::new(&name, &root_output, &build_targets, &install_paths);
+    let mut finger_print =
+        FingerPrint::new(&name, &root_output, build_targets.clone(), &install_paths);
 
     let pristine = finger_print.load_previous().is_err();
 
@@ -785,7 +978,7 @@ pub fn cbuild(
     }
 
     let exec = Arc::new(Exec::default());
-    let _r = compile_with_exec(
+    let out_dir = compile_with_exec(
         ws,
         &compile_opts,
         &(exec.clone() as Arc<dyn Executor>),
@@ -793,6 +986,18 @@ pub fn cbuild(
         &capi_config.library.rustflags,
     )?;
 
+    build_targets
+        .extra
+        .setup(&capi_config, &root_path, out_dir.as_deref())?;
+
+    if capi_config.header.generation {
+        let mut header_name = PathBuf::from(&capi_config.header.name);
+        header_name.set_extension("h");
+        let from = root_output.join(&header_name);
+        let to = Path::new(&capi_config.header.subdirectory).join(&header_name);
+        build_targets.extra.include.push((from, to));
+    }
+
     if pristine {
         // restore the default to make sure the tests do not trigger a second rebuild.
         compile_opts.build_config.force_rebuild = false;
@@ -812,10 +1017,11 @@ pub fn cbuild(
         }
         pc.add_lib_private(&static_libs);
 
-        build_pc_files(&ws, &name, &root_output, &pc)?;
+        build_pc_files(&ws, &capi_config.pkg_config.filename, &root_output, &pc)?;
 
         if !only_staticlib {
-            build_def_file(&ws, &name, &rustc_target, &root_output)?;
+            let lib_name = name;
+            build_def_file(&ws, &lib_name, &rustc_target, &root_output)?;
 
             let mut dlltool = std::env::var_os("DLLTOOL")
                 .map(PathBuf::from)
@@ -826,27 +1032,38 @@ pub fn cbuild(
                 dlltool = args.value_of("dlltool").map(PathBuf::from).unwrap();
             }
 
-            build_implib_file(&ws, &name, &rustc_target, &root_output, &dlltool)?;
+            build_implib_file(&ws, &lib_name, &rustc_target, &root_output, &dlltool)?;
         }
 
         if capi_config.header.enabled {
             let header_name = &capi_config.header.name;
             if capi_config.header.generation {
                 build_include_file(&ws, header_name, &version, &root_output, &root_path)?;
-            } else {
-                copy_prebuilt_include_file(&ws, header_name, &root_output, &root_path)?;
             }
+
+            copy_prebuilt_include_file(&ws, &build_targets, &root_output)?;
         }
 
-        // Generate versioned links in target dir so the lib can be used uninstalled
-        if let Some(ref shared_lib) = build_targets.shared_lib {
-            if capi_config.library.versioning {
-                let lib_name = &capi_config.library.name;
-                let lib_type = LibType::from_build_targets(&build_targets);
-                let lib = UnixLibNames::new(lib_type, lib_name, &capi_config.library.version);
-                if let Some(lib) = lib {
-                    lib.install(&capi_config, &shared_lib, &root_output)?;
-                }
+        if name.contains('-') {
+            let from_build_targets = BuildTargets::new(
+                &name.replace("-", "_"),
+                &rustc_target,
+                &root_output,
+                &libkinds,
+                &capi_config,
+            )?;
+
+            if let (Some(from_static_lib), Some(to_static_lib)) = (
+                from_build_targets.static_lib.as_ref(),
+                build_targets.static_lib.as_ref(),
+            ) {
+                copy(from_static_lib, to_static_lib)?;
+            }
+            if let (Some(from_shared_lib), Some(to_shared_lib)) = (
+                from_build_targets.shared_lib.as_ref(),
+                build_targets.shared_lib.as_ref(),
+            ) {
+                copy(from_shared_lib, to_shared_lib)?;
             }
         }
 
